@@ -1,5 +1,5 @@
 /* remove.c -- core functions for removing files and directories
-   Copyright (C) 88, 90, 91, 1994-2006 Free Software Foundation, Inc.
+   Copyright (C) 88, 90, 91, 1994-2007 Free Software Foundation, Inc.
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -148,16 +148,6 @@ struct dirstack_state
 };
 typedef struct dirstack_state Dirstack_state;
 
-/* Just like close(fd), but don't modify errno. */
-static inline int
-close_preserve_errno (int fd)
-{
-  int saved_errno = errno;
-  int result = close (fd);
-  errno = saved_errno;
-  return result;
-}
-
 /* Like fstatat, but cache the result.  If ST->st_size is -1, the
    status has not been gotten yet.  If less than -1, fstatat failed
    with errno == -1 - ST->st_size.  Otherwise, the status has already
@@ -173,11 +163,12 @@ cache_fstatat (int fd, char const *file, struct stat *st, int flag)
   return -1;
 }
 
-/* Initialize a fstatat cache *ST.  */
-static inline void
+/* Initialize a fstatat cache *ST.  Return ST for convenience.  */
+static inline struct stat *
 cache_stat_init (struct stat *st)
 {
   st->st_size = -1;
+  return st;
 }
 
 /* Return true if *ST has been statted.  */
@@ -426,22 +417,32 @@ ds_free (Dirstack_state *ds)
   free (ds);
 }
 
-/* Pop the active directory (AD) stack and move *DIRP `up' one level,
+/* Pop the active directory (AD) stack and prepare to move `up' one level,
    safely.  Moving `up' usually means opening `..', but when we've just
    finished recursively processing a command-line directory argument,
-   there's nothing left on the stack, so set *DIRP to NULL in that case.
-   The idea is to return with *DIRP opened on the parent directory,
+   there's nothing left on the stack, so set *FDP to AT_FDCWD in that case.
+   The idea is to return with *FDP opened on the parent directory,
    assuming there are entries in that directory that we need to remove.
 
+   Note that we must not call opendir (or fdopendir) just yet, since
+   the caller must first remove the directory we're coming from.
+   That is because some file system implementations cache readdir
+   results at opendir time; so calling opendir, rmdir, readdir would
+   return an entry for the just-removed directory.
+
    Whenever using chdir '..' (virtually, now, via openat), verify
    that the post-chdir dev/ino numbers for `.' match the saved ones.
-   If any system call fails or if dev/ino don't match then give a
+   If any system call fails or if dev/ino don't match, then give a
    diagnostic and longjump out.
-   Set *PREV_DIR to the name (in malloc'd storage) of the
+   Return the name (in malloc'd storage) of the
    directory (usually now empty) from which we're coming, and which
-   corresponds to the input value of *DIRP.  */
-static void
-AD_pop_and_chdir (DIR **dirp, Dirstack_state *ds, char **prev_dir)
+   corresponds to the input value of DIRP.
+
+   Finally, note that while this function's name is no longer as
+   accurate as it once was (it no longer calls chdir), it does open
+   the destination directory.  */
+static char *
+AD_pop_and_chdir (DIR *dirp, int *fdp, Dirstack_state *ds)
 {
   struct AD_ent *leaf_dir_ent = AD_stack_top(ds);
   struct dev_ino leaf_dev_ino = leaf_dir_ent->dev_ino;
@@ -450,7 +451,7 @@ AD_pop_and_chdir (DIR **dirp, Dirstack_state *ds, char **prev_dir)
 
   /* Get the name of the current (but soon to be `previous') directory
      from the top of the stack.  */
-  *prev_dir = top_dir (ds);
+  char *prev_dir = top_dir (ds);
 
   AD_stack_pop (ds);
   pop_dir (ds);
@@ -474,15 +475,15 @@ AD_pop_and_chdir (DIR **dirp, Dirstack_state *ds, char **prev_dir)
   if (1 < AD_stack_height (ds))
     {
       struct stat sb;
-      int fd = openat (dirfd (*dirp), "..", O_RDONLY);
-      if (closedir (*dirp) != 0)
+      int fd = openat (dirfd (dirp), "..", O_RDONLY);
+      if (closedir (dirp) != 0)
 	{
 	  error (0, errno, _("FATAL: failed to close directory %s"),
-		 quote (full_filename (*prev_dir)));
+		 quote (full_filename (prev_dir)));
 	  goto next_cmdline_arg;
 	}
 
-      /* The above fails with EACCES when *DIRP is readable but not
+      /* The above fails with EACCES when DIRP is readable but not
 	 searchable, when using Solaris' openat.  Without this openat
 	 call, tests/rm2 would fail to remove directories a/2 and a/3.  */
       if (fd < 0)
@@ -491,7 +492,7 @@ AD_pop_and_chdir (DIR **dirp, Dirstack_state *ds, char **prev_dir)
       if (fd < 0)
 	{
 	  error (0, errno, _("FATAL: cannot open .. from %s"),
-		 quote (full_filename (*prev_dir)));
+		 quote (full_filename (prev_dir)));
 	  goto next_cmdline_arg;
 	}
 
@@ -508,39 +509,32 @@ AD_pop_and_chdir (DIR **dirp, Dirstack_state *ds, char **prev_dir)
 	{
 	  error (0, 0, _("FATAL: directory %s changed dev/ino"),
 		 quote (full_filename (".")));
-	  goto close_and_next;
-	}
-
-      *dirp = fdopendir (fd);
-      if (*dirp == NULL)
-	{
-	  error (0, errno, _("FATAL: cannot return to .. from %s"),
-		 quote (full_filename (".")));
-
 	close_and_next:;
 	  close (fd);
 
 	next_cmdline_arg:;
-	  free (*prev_dir);
+	  free (prev_dir);
 	  longjmp (ds->current_arg_jumpbuf, 1);
 	}
+      *fdp = fd;
     }
   else
     {
-      if (closedir (*dirp) != 0)
+      if (closedir (dirp) != 0)
 	{
 	  error (0, errno, _("FATAL: failed to close directory %s"),
-		 quote (full_filename (*prev_dir)));
+		 quote (full_filename (prev_dir)));
 	  goto next_cmdline_arg;
 	}
-      *dirp = NULL;
+      *fdp = AT_FDCWD;
     }
+
+  return prev_dir;
 }
 
-/* Initialize *HT if it is NULL.
-   Insert FILENAME into HT.  */
-static void
-AD_mark_helper (Hash_table **ht, char *filename)
+/* Initialize *HT if it is NULL.  Return *HT.  */
+static Hash_table *
+AD_ensure_initialized (Hash_table **ht)
 {
   if (*ht == NULL)
     {
@@ -549,7 +543,16 @@ AD_mark_helper (Hash_table **ht, char *filename)
       if (*ht == NULL)
 	xalloc_die ();
     }
-  void *ent = hash_insert (*ht, filename);
+
+  return *ht;
+}
+
+/* Initialize *HT if it is NULL.
+   Insert FILENAME into HT.  */
+static void
+AD_mark_helper (Hash_table **ht, char *filename)
+{
+  void *ent = hash_insert (AD_ensure_initialized (ht), filename);
   if (ent == NULL)
     xalloc_die ();
   else
@@ -557,7 +560,6 @@ AD_mark_helper (Hash_table **ht, char *filename)
       if (ent != filename)
 	free (filename);
     }
-
 }
 
 /* Mark FILENAME (in current directory) as unremovable.  */
@@ -796,10 +798,14 @@ prompt (int fd_cwd, Dirstack_state const *ds, char const *filename,
 
   *is_empty = T_UNKNOWN;
 
-  if (((!x->ignore_missing_files & (x->interactive | x->stdin_tty))
+  if (x->interactive == RMI_NEVER)
+    return RM_OK;
+
+  if (((!x->ignore_missing_files & ((x->interactive == RMI_ALWAYS)
+				    | x->stdin_tty))
        && (write_protected = write_protected_non_symlink (fd_cwd, filename,
 							  ds, sbuf)))
-      || x->interactive)
+      || x->interactive == RMI_ALWAYS)
     {
       if (cache_fstatat (fd_cwd, filename, sbuf, AT_SYMLINK_NOFOLLOW) != 0)
 	{
@@ -819,7 +825,7 @@ prompt (int fd_cwd, Dirstack_state const *ds, char const *filename,
       /* Using permissions doesn't make sense for symlinks.  */
       if (S_ISLNK (sbuf->st_mode))
 	{
-	  if ( ! x->interactive)
+	  if (x->interactive != RMI_ALWAYS)
 	    return RM_OK;
 	  write_protected = false;
 	}
@@ -1016,8 +1022,7 @@ remove_entry (int fd_cwd, Dirstack_state const *ds, char const *filename,
 	errno = EISDIR;
 
       if (! x->recursive
-	  || errno == ENOENT || errno == ENOTDIR
-	  || errno == ELOOP || errno == ENAMETOOLONG)
+	  || (cache_stat_ok (st) && !S_ISDIR (st->st_mode)))
 	{
 	  if (ignorable_missing (x, errno))
 	    return RM_OK;
@@ -1028,6 +1033,7 @@ remove_entry (int fd_cwd, Dirstack_state const *ds, char const *filename,
 		 quote (full_filename (filename)));
 	  return RM_ERROR;
 	}
+      assert (!cache_stat_ok (st) || S_ISDIR (st->st_mode));
     }
   else
     {
@@ -1106,37 +1112,33 @@ remove_entry (int fd_cwd, Dirstack_state const *ds, char const *filename,
 static DIR *
 fd_to_subdirp (int fd_cwd, char const *f,
 	       struct rm_options const *x, int prev_errno,
-	       struct stat *subdir_sb, Dirstack_state *ds,
+	       struct stat *subdir_sb,
 	       int *cwd_errno ATTRIBUTE_UNUSED)
 {
   int open_flags = O_RDONLY | O_NOCTTY | O_NOFOLLOW | O_NONBLOCK;
   int fd_sub = openat_permissive (fd_cwd, f, open_flags, 0, cwd_errno);
+  int saved_errno;
 
   /* Record dev/ino of F.  We may compare them against saved values
      to thwart any attempt to subvert the traversal.  They are also used
      to detect directory cycles.  */
-  if (fd_sub < 0 || fstat (fd_sub, subdir_sb) != 0)
+  if (fd_sub < 0)
+    return NULL;
+  else if (fstat (fd_sub, subdir_sb) != 0)
+    saved_errno = errno;
+  else if (S_ISDIR (subdir_sb->st_mode))
     {
-      if (0 <= fd_sub)
-	close_preserve_errno (fd_sub);
-      return NULL;
+      DIR *subdir_dirp = fdopendir (fd_sub);
+      if (subdir_dirp)
+	return subdir_dirp;
+      saved_errno = errno;
     }
+  else
+    saved_errno = (prev_errno ? prev_errno : ENOTDIR);
 
-  if (! S_ISDIR (subdir_sb->st_mode))
-    {
-      errno = prev_errno ? prev_errno : ENOTDIR;
-      close_preserve_errno (fd_sub);
-      return NULL;
-    }
-
-  DIR *subdir_dirp = fdopendir (fd_sub);
-  if (subdir_dirp == NULL)
-    {
-      close_preserve_errno (fd_sub);
-      return NULL;
-    }
-
-  return subdir_dirp;
+  close (fd_sub);
+  errno = saved_errno;
+  return NULL;
 }
 
 /* Remove entries in the directory open on DIRP
@@ -1220,7 +1222,7 @@ remove_cwd_entries (DIR **dirp,
 	case RM_NONEMPTY_DIR:
 	  {
 	    DIR *subdir_dirp = fd_to_subdirp (dirfd (*dirp), f,
-					      x, errno, subdir_sb, ds, NULL);
+					      x, errno, subdir_sb, NULL);
 	    if (subdir_dirp == NULL)
 	      {
 		status = RM_ERROR;
@@ -1298,6 +1300,7 @@ remove_dir (int fd_cwd, Dirstack_state *ds, char const *dir,
 	    struct rm_options const *x, int *cwd_errno)
 {
   enum RM_status status;
+  dev_t current_dev = dir_st->st_dev;
 
   /* There is a race condition in that an attacker could replace the nonempty
      directory, DIR, with a symlink between the preceding call to rmdir
@@ -1307,7 +1310,7 @@ remove_dir (int fd_cwd, Dirstack_state *ds, char const *dir,
      fd_to_subdirp's fstat, along with the `fstat' and the dev/ino
      comparison in AD_push ensure that we detect it and fail.  */
 
-  DIR *dirp = fd_to_subdirp (fd_cwd, dir, x, 0, dir_st, ds, cwd_errno);
+  DIR *dirp = fd_to_subdirp (fd_cwd, dir, x, 0, dir_st, cwd_errno);
 
   if (dirp == NULL)
     {
@@ -1359,23 +1362,38 @@ remove_dir (int fd_cwd, Dirstack_state *ds, char const *dir,
 	}
       if (subdir)
 	{
-	  AD_push (dirfd (dirp), ds, subdir, &subdir_sb);
-	  AD_INIT_OTHER_MEMBERS ();
+	  if ( ! x->one_file_system
+	       || subdir_sb.st_dev == current_dev)
+	    {
+	      AD_push (dirfd (dirp), ds, subdir, &subdir_sb);
+	      AD_INIT_OTHER_MEMBERS ();
+	      free (subdir);
+	      continue;
+	    }
 
+	  /* Here, --one-file-system is in effect, and with remove_cwd_entries'
+	     traversal into the current directory, (known as SUBDIR, from ..),
+	     DIRP's device number is different from CURRENT_DEV.  Arrange not
+	     to do anything more with this hierarchy.  */
+	  error (0, 0, _("skipping %s, since it's on a different device"),
+		 quote (full_filename (subdir)));
 	  free (subdir);
-	  continue;
+	  AD_mark_current_as_unremovable (ds);
+	  tmp_status = RM_ERROR;
+	  UPDATE_STATUS (status, tmp_status);
 	}
 
       /* Execution reaches this point when we've removed the last
-	 removable entry from the current directory.  */
+	 removable entry from the current directory -- or, with
+	 --one-file-system, when the current directory is on a
+	 different file system.  */
       {
+	int fd;
 	/* The name of the directory that we have just processed,
 	   nominally removing all of its contents.  */
-	char *empty_dir;
-
-	AD_pop_and_chdir (&dirp, ds, &empty_dir);
-	int fd = (dirp != NULL ? dirfd (dirp) : AT_FDCWD);
-	assert (dirp != NULL || AD_stack_height (ds) == 1);
+	char *empty_dir = AD_pop_and_chdir (dirp, &fd, ds);
+	dirp = NULL;
+	assert (fd != AT_FDCWD || AD_stack_height (ds) == 1);
 
 	/* Try to remove EMPTY_DIR only if remove_cwd_entries succeeded.  */
 	if (tmp_status == RM_OK)
@@ -1386,14 +1404,16 @@ remove_dir (int fd_cwd, Dirstack_state *ds, char const *dir,
 	       But that's no big deal since we're interactive.  */
 	    struct stat empty_st;
 	    Ternary is_empty;
-	    cache_stat_init (&empty_st);
-	    enum RM_status s = prompt (fd, ds, empty_dir, &empty_st, x,
+	    enum RM_status s = prompt (fd, ds, empty_dir,
+				       cache_stat_init (&empty_st), x,
 				       PA_REMOVE_DIR, &is_empty);
 
 	    if (s != RM_OK)
 	      {
 		free (empty_dir);
 		status = s;
+		if (fd != AT_FDCWD)
+		  close (fd);
 		goto closedir_and_return;
 	      }
 
@@ -1415,8 +1435,17 @@ remove_dir (int fd_cwd, Dirstack_state *ds, char const *dir,
 
 	free (empty_dir);
 
-	if (AD_stack_height (ds) == 1)
+	if (fd == AT_FDCWD)
 	  break;
+
+	dirp = fdopendir (fd);
+	if (dirp == NULL)
+	  {
+	    error (0, errno, _("FATAL: cannot return to .. from %s"),
+		   quote (full_filename (".")));
+	    close (fd);
+	    longjmp (ds->current_arg_jumpbuf, 1);
+	  }
       }
     }
 
@@ -1454,6 +1483,7 @@ rm_1 (Dirstack_state *ds, char const *filename,
 
   struct stat st;
   cache_stat_init (&st);
+  cycle_check_init (&ds->cycle_check_state);
   if (x->root_dev_ino)
     {
       if (cache_fstatat (AT_FDCWD, filename, &st, AT_SYMLINK_NOFOLLOW) != 0)
@@ -1473,8 +1503,7 @@ rm_1 (Dirstack_state *ds, char const *filename,
   AD_push_initial (ds);
   AD_INIT_OTHER_MEMBERS ();
 
-  int fd_cwd = AT_FDCWD;
-  enum RM_status status = remove_entry (fd_cwd, ds, filename, &st, x, NULL);
+  enum RM_status status = remove_entry (AT_FDCWD, ds, filename, &st, x, NULL);
   if (status == RM_NONEMPTY_DIR)
     {
       /* In the event that remove_dir->remove_cwd_entries detects
@@ -1483,13 +1512,12 @@ rm_1 (Dirstack_state *ds, char const *filename,
       if (setjmp (ds->current_arg_jumpbuf))
 	status = RM_ERROR;
       else
-	status = remove_dir (fd_cwd, ds, filename, &st, x, cwd_errno);
+	status = remove_dir (AT_FDCWD, ds, filename, &st, x, cwd_errno);
 
       AD_stack_clear (ds);
     }
 
   ds_clear (ds);
-
   return status;
 }
 
@@ -1509,13 +1537,13 @@ rm (size_t n_files, char const *const *file, struct rm_options const *x)
 	{
 	  error (0, 0, _("cannot remove relative-named %s"), quote (file[i]));
 	  status = RM_ERROR;
-	  continue;
 	}
-
-      cycle_check_init (&ds->cycle_check_state);
-      enum RM_status s = rm_1 (ds, file[i], x, &cwd_errno);
-      assert (VALID_STATUS (s));
-      UPDATE_STATUS (status, s);
+      else
+	{
+	  enum RM_status s = rm_1 (ds, file[i], x, &cwd_errno);
+	  assert (VALID_STATUS (s));
+	  UPDATE_STATUS (status, s);
+	}
     }
 
   if (x->require_restore_cwd && cwd_errno)
